/*
 * fat_dirent.cpp:
 * Implements the FAT directory entry class
 */

#include "ffs.h"
#include <string.h>
#include <ctype.h>

/****************************************************************
 ****************************************************************
 ****************************************************************/

fat_dirent::fat_dirent(class fat_BPB *bpb_)
{
    bpb = bpb_;
    init();
}

void fat_dirent::init(void)
{
    memset(namebuf,0,sizeof(namebuf));
    memset(namebuf2,0,sizeof(namebuf2));
    memset(DIR_Name,0,sizeof(DIR_Name));
    memset(DIR_unicode,0,sizeof(DIR_unicode));
    DIR_unicode_len = 0;
    DIR_CrtTimeTenth = 0;
    memset(&atime,0,sizeof(atime));
    memset(&mtime,0,sizeof(mtime));
    memset(&ctime,0,sizeof(ctime));
    DIR_FstClus = 0;
    DIR_FileSize = 0;
    deleted = false;
    valcode = ZERO;
    complete = false;
    memset(attrbuf,0,sizeof(attrbuf));
    memset(datebuf,0,sizeof(datebuf));
}

fat_dirent::~fat_dirent()
{
    valcode = UNDEFINED;
}

void fat_dirent::dosdate2tm(const unsigned char *buf,struct tm *tm)
{
    // bits 0-4, day of month, 1-31

    tm->tm_mday = buf[0] & 0x1f;	

    // bits 5-8: months of year, 1-12

    tm->tm_mon  = (((buf[0] & 0xe0) >> 5) |
		   ((buf[1] & 0x01) << 3)) - 1; 

    // bits 9-15, years since 1980, 0-127

    tm->tm_year = ((buf[1] & 0xfe) >> 1) + 80;	
}

void fat_dirent::dostime2tm(const unsigned char *buf,struct tm *tm)
{
    /* First assemble the 16-bit value */
    unsigned t = buf[1] << 8 | buf[0];

    /* Now decode the bits */

    // bits 0-4, 2-second count, 0-29
    tm->tm_sec  = (t & 0x1f) * 2;	

    // bits 5-10, minutes, range 0-59
    tm->tm_min  = (t & 0x7e0) >> 5;

    // bits 11-15, hours, range 0-23
    tm->tm_hour = (t & 0xf800) >> 11;
}

bool fat_dirent::tmvalid(struct tm *tm)
{
    /* Special-case code for DOS losing values */
    if(tm->tm_year==80 && tm->tm_mon==-1 &&
       tm->tm_mday==0  && tm->tm_hour==0 &&
       tm->tm_min==0 && tm->tm_sec==0){
	return true;
    }

    if(tm->tm_mday!=0 && tm->tm_mon!= -1){
	if(tm->tm_mday < 1  || tm->tm_mday>31) return false;
	if(tm->tm_mon < 0   || tm->tm_mon > 11) return false;
    }
    if(tm->tm_year < 80 || tm->tm_year > 207) return false;
    if(tm->tm_sec < 0   || tm->tm_sec > 59) return false;
    if(tm->tm_min < 0   || tm->tm_min > 59) return false;
    if(tm->tm_hour < 0  || tm->tm_hour > 23) return false;

    /* Be a little more restrictive on year */
    if(tm->tm_year > 110){		// > 2010
	return false;
    }

    return true;
}

/* Read a 32-byte unicode directory buffer */
bool fat_dirent::readbuf_LDIR(const uchar *buf)
{
    valcode = UNDEFINED;

    if(buf[11] != ATTR_LONG_NAME) return false;

    if(buf[12] != 0){
	//fprintf(stderr,"buf[12] is %x (should be 0)\n",buf[12]);
	return false;
    }
    if(buf[26] != 0){
	//fprintf(stderr,"buf[26] is bad\n");
	return false;
    }
    if(buf[27] != 0){
	//fprintf(stderr,"buf[27] is bad\n");
	return false;
    }

    /* Clear the FAT16 values that make no sense */
    memset(DIR_Name,0,sizeof(DIR_Name));
    DIR_FstClus   = 0;
    DIR_FileSize  = 0;
    DIR_NTRes     = 0;
    DIR_CrtTimeTenth = 0;
    memset(&ctime,0,sizeof(ctime));
    memset(&atime,0,sizeof(atime));
    memset(&mtime,0,sizeof(mtime));
    
    /* Set the values that make sense */

    LDIR_Ord  = (unsigned char)(buf[0] & ~LAST_LONG_ENTRY);
    LDIR_last = buf[0] & LAST_LONG_ENTRY ? true : false;
    if(LDIR_Ord==1){
	LDIR_last = true;
    }
    LDIR_Chksum = buf[13];

    /* This seems to be set on deleted names */
    deleted   = (LDIR_Ord == 165);


    /* Get the name */

    memcpy(LDIR_unicode,buf+1,10);	// characters 1-5
    memcpy(LDIR_unicode+10,buf+14,12);	// characters 6-11
    memcpy(LDIR_unicode+22,buf+28,4);	// character 12-13

    memmove(DIR_unicode+26,DIR_unicode,DIR_unicode_len);
    DIR_unicode_len += 26;
    memmove(DIR_unicode,LDIR_unicode,26);

    valcode = LDIR_ONLY;

    for(int k=0;k<DIR_unicode_len;k+=2){
	if(DIR_unicode[k]==255){
	    valcode = LDIR_TERM;
	}
    }
    return true;
}

/* Read a 32-byte directory buffer */
bool fat_dirent::readbuf(const uchar *buf)
{
    const unsigned char *fname = buf;
    int i;


    LDIR	= false;
    valcode     = UNDEFINED;

    DIR_Attr = buf[11];
    if(DIR_Attr==ATTR_LONG_NAME) return readbuf_LDIR(buf);

    DIR_CrtTimeTenth = buf[13];

    if(DIR_Attr > 0x2f){
	valcode = BAD_ATTR;
	return false;	// invalid attribute bit
    }

    if(DIR_CrtTimeTenth > 199){
	valcode = BAD_CRTTIME;
	return false; // out of range
    }

    /* validate fname */
    int nonspace = 0;

    /* check for . directories I've seen */
    if(fname[0]==0x2e && fname[1]==' ' && fname[2]==' ' && fname[3]==' ' &&
       fname[4]==' ' && fname[5]==' ' && fname[6]==' ' && fname[7]==' ' &&
       fname[8]==' ' && fname[9]==' ' && fname[10]==' ' && fname[11]=='\020'){
	goto next;
    }

    for(i=0;i<11;i++){
	if(i==0 && (fname[i]==0 || fname[i]==0xe5 || fname[i]==0x05)){
	    continue;			// all valid in first position
	}
	if(fname[i] < 0x20){
	    valcode = BAD_NAME;
	    return false;			// not valid here
	}
	if(fname[i]==0x22 || fname[i]==0x2a || fname[i]==0x2b ||
	   fname[i]==0x2c || fname[i]==0x2e || fname[i]==0x2f ||
	   fname[i]==0x3a || fname[i]==0x3b || fname[i]==0x3c || 
	   fname[i]==0x3d || fname[i]==0x3e || fname[i]==0x3f || 
	   fname[i]==0x5b || fname[i]==0x5c || fname[i]==0x5d || 
	   fname[i]==0x7c){
	    valcode = BAD_NAME;
	    return false;			// invalid character
	}
	if(!isspace(fname[i])) nonspace++;
    }
    if(nonspace==0){			// must be one non-blank char
	valcode = BAD_NAME;
	return false;
    }
 next:;
    
    memcpy(DIR_Name,fname,11);
    DIR_Name[11] = 0;
    if(DIR_Name[0]==0 || DIR_Name[0]==0xe5){
	DIR_Name[0] = '?';
	deleted = true;
	/* If we have a unicode name, use it's first character. */
	if(DIR_unicode_len>0){
	    DIR_Name[0] =  DIR_unicode[0];
	    if(islower(DIR_Name[0])) DIR_Name[0] = (uchar)toupper(DIR_Name[0]);
	}
    }
    if(DIR_Name[0]==0x05){
	DIR_Name[0] = (unsigned char)0xe5;		// kanji!
    }

    memset(&ctime,0,sizeof(ctime));
    memset(&atime,0,sizeof(atime));
    memset(&mtime,0,sizeof(mtime));

    dostime2tm(buf+14,&ctime);
    dosdate2tm(buf+16,&ctime);
    if(tmvalid(&ctime)==false){		// ctime not valid
	valcode = BAD_CTIME;
	return false;
    }

    dosdate2tm(buf+18,&atime);
    if(tmvalid(&atime)==false){		// atime not valid
	valcode = BAD_ATIME;
	return false;
    }

    dostime2tm(buf+22,&mtime);
    dosdate2tm(buf+24,&mtime);
    if(tmvalid(&mtime)==false){		// mtime not valid
	valcode = BAD_MTIME;
	return false;
    }

    /* Assume that the maximum number of clusters and filesize are
     * big, unless we know otherwise
     */
    ulong max_clusters = 8388608; // 4GB drive, single-sector clusters (2^32 / 512)
    ulong max_filesec  = max_clusters / 2; // 2GB max filesize, in sectors

    if(bpb){
	if(bpb->CountofClusters) max_clusters = bpb->CountofClusters;
	if(bpb->TotSec)          max_filesec  = bpb->TotSec;
    }

    DIR_FstClus  = (intel::u2(buf+20)<<16) | intel::u2(buf+26);
    if(DIR_FstClus > max_clusters ){
	valcode = BAD_FSTCLUS;
	return false;
    }
	
    DIR_FileSize = intel::u4(buf+28);		// file size
    if(DIR_FileSize/512 > max_filesec){
	valcode = BAD_FILESIZE;
	return false;
    }

    valcode = VALID;
    complete = true;
    return true;		// would be nice to validate the times
}

const char *fat_dirent::mtime2ascii()
{
    if(datebuf[0]==0){
	sprintf(datebuf,"%02d/%02d/%04d  %02d:%02d  ",
		mtime.tm_mon+1,
		mtime.tm_mday,
		mtime.tm_year+1900,
		mtime.tm_hour,
		mtime.tm_min);
    }
    return datebuf;
}


const	char *fat_dirent::attr2ascii()
{
    if(attrbuf[0]==0){
	attrbuf[0] = DIR_Attr & fat_dirent::ATTR_READ_ONLY ? 'R' : ' ';
	attrbuf[1] = DIR_Attr & fat_dirent::ATTR_HIDDEN ? 'H' : ' ';
	attrbuf[2] = DIR_Attr & fat_dirent::ATTR_SYSTEM ? 'S' : ' ';
	attrbuf[3] = DIR_Attr & fat_dirent::ATTR_ARCHIVE ? 'A' : ' ';
	attrbuf[4] = DIR_Attr & fat_dirent::ATTR_VOLUME_ID ? 'V' : ' ';
	attrbuf[5] = 0;
    }
    return attrbuf;
}

void	fat_dirent::fwrite(FILE *f,class img *img)
{
    int present = 0;

    if(img){
	present = sector() < img->image_blocks;
    }
    
    /* IF we are just a LDIR entry, print that but nothing else */
    if(LDIR){
	fprintf(f,"   LDIR [%d]:     ",LDIR_Ord);
	for(int k=0;k<26;k+=2){
	    fprintf(f,"%c",LDIR_unicode[k]);
	}
	fputc('\n',f);

	return;
    }

    if(DIR_FstClus==0) present=0; // not present if cluster is 0
    
    bool print_partial = false;

    if((DIR_Attr&fat_dirent::ATTR_LONG_NAME)==fat_dirent::ATTR_LONG_NAME){
	/* We were asked to print a directory entry, but all we have is the
	 * partial file name. Just print what we have and returned; that's all we know.
	 */
	fprintf(f,"%19s[LDIR %2d] %12s","",LDIR_Ord,"");
	print_partial=true;
    }
    else{
	fputs(mtime2ascii(),f);
	fputs(attr2ascii(),f);
    }

    if(print_partial==false){
	fprintf(f,"%5s ",is_dir() ? "<DIR>" :"");
	char buf[64];
	fprintf(f,"%10s ", ffs::itocomma(buf,DIR_FileSize));
    }

    
    /* If we have a long directory name, print that, then print the regular directory
     * name on the next line, unless this is a deleted entry.... In that case, just print
     * it.
     */
    if(DIR_unicode_len>0){
	fprintf(f,"%s",name());

	if(print_partial && DIR_FstClus==0){
	    fputc('\n',f);
	    return;				// alas, not very satisfying
	}

	if(deleted==false || LDIR_last){
	    fputc('\n',f);
	    fprintf(f,"%40s ( %s )","",DIR_Name);
	}
	else {
	    fprintf(f,":del");
	}
    }
    else{
	fprintf(f,"%s%4s", DIR_Name,deleted ? ":del" : "");
    }

    fputc(' ',f);
    if(present){
	fprintf(f,"+cluster %d+ ",DIR_FstClus);
    }
    else {
	fprintf(f,"  (cluster %6d / sector %6d)",DIR_FstClus,sector());
    }
    fputc('\n',f);
}

sector_t fat_dirent::sector()
{
    return bpb ? bpb->cluster2sector(DIR_FstClus) : 0;
}

bool fat_dirent::is_null()
{
    return
	DIR_Name[0] == 0     &&
	DIR_unicode_len == 0 &&
	DIR_Attr == 0 &&
	DIR_CrtTimeTenth == 0 &&
	mtime.tm_year==0 &&
	mtime.tm_mon==0 &&
	mtime.tm_mday==0 &&
	mtime.tm_hour==0 &&
	mtime.tm_min==0 &&
	mtime.tm_sec==0 &&
	DIR_FstClus == 0 &&
	DIR_FileSize==0;
}

bool fat_dirent::is_dir()
{
    return DIR_Attr & fat_dirent::ATTR_DIRECTORY;
}

bool fat_dirent::is_complete()
{
    return complete;
}

bool fat_dirent::similar(class fat_dirent &d1)
{
    int i;
    for(i=0;i<sizeof(DIR_Name);i++){
	if(DIR_Name[i] != d1.DIR_Name[i]) return false;	// name is different
    }
    if(DIR_FstClus != d1.DIR_FstClus) return false;
    if(DIR_FileSize != d1.DIR_FileSize) return false;
    if(memcmp(&mtime,&d1.mtime,sizeof(mtime))) return false;
    return true;
}

int fat_dirent::level()
{
    if(deleted) return 2;		// by definition
    if(ffs::analyze_islevel1(name())) return 1;	// we look like level1 data
    if(name_alias() && ffs::analyze_islevel1(name_alias())){
	return 1;
    }
    return 0;
}

/* delname() - a unique suffix for deleted names */
const char *fat_dirent::delname()
{
    sprintf(delnamebuf,":del%d",DIR_FstClus);
    return delnamebuf;
}

/* Save a trimmed 8.3 name to buf */
void fat_dirent::save83name(char buf[32])
{
    int i;

    for(i=0;i<8;i++){
	buf[i] = DIR_Name[i];
    }
    for(i=7;i>0;i--){
	if(buf[i]==' ') buf[i]='\000';
	else break;
    }
    for(i=0;i<8;i++){
	if(buf[i]=='\000') break;
    }
    /* i now points to the null at the end of the string */
    if(DIR_Name[8]!=' '){
	buf[i++] = '.';
	buf[i++] = DIR_Name[8];
	if(DIR_Name[9]!=' '){
	    buf[i++] = DIR_Name[9];
	    if(DIR_Name[10]!=' '){
		buf[i++] = DIR_Name[10];
	    }
	}
    }
    buf[i]='\000';
    if(deleted) strcat(buf,delname());
}

/* name() - a name that is easy to use */
const char *fat_dirent::name()
{
    int i,k;
    if(namebuf[0]) return namebuf;

    /* If we have a UNICODE name, use that */
    if(DIR_unicode_len>0){
	for(i=0,k=0;k<DIR_unicode_len;i++,k+=2){
	    namebuf[i] = DIR_unicode[k]!=255 ? DIR_unicode[k] : ' ';
	}
	namebuf[i] = '\000';
	if(deleted) strcat(namebuf,delname());
	return namebuf;
    }

    /* Otherwise, return the 8.3 name, trimmed and made 8.3 */
    save83name(namebuf);
    return namebuf;
}

const char *fat_dirent::name_alias()
{
    if(DIR_unicode[0]) return 0;	// if no unicode, then there is no name alias
    save83name(namebuf2);
    namebuf2[sizeof(namebuf2)-1] = 0;	// make sure it is null-terminated
    return namebuf2;
}
