/* gui.cpp:
 * Graphical user interface for fatdump
 * Mostly optimized for Unix right now.
 */

#include "StdAfx.h"
#include "ffs.h"
#include "gui.h"

#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
#include <io.h>
#include <conio.h>
#endif

#ifdef UNIX
#include <curses.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <unistd.h>
#endif


#ifdef WIN32
int keychar()
{
    int c = getch();
    if(c==0x03){
	fprintf(stderr,"^C\n");
	exit(1);
    }
    if(c>=' ') putchar(c);
    return c;
}
#else
#include <sgtty.h>
int	keychar()
{
	struct	sgttyb	sg_old,sg_new;
	char	ret;

	fflush(stdout);				
	ioctl(fileno(stdin),TIOCGETP,&sg_old);

	sg_new		= sg_old;		/* make a copy */
	sg_new.sg_flags	|= CBREAK;		/* turn cbreak on */
	sg_new.sg_flags &= ~ECHO;		/* turn echo off */

	ioctl(fileno(stdin),TIOCSETP,&sg_new);

	read(fileno(stdin),&ret,1);

	ioctl(fileno(stdin),TIOCSETP,&sg_old);	/* restore old mode */
	if(ret==0x03){
	    fprintf(stderr,"^C\n");
	    exit(1);
	}
	putchar(ret);
	return(ret);
}
int	kbhit()
{
    struct timeval timeout;

    memset(&timeout,0,sizeof(timeout));

    fd_set rfds;
    FD_ZERO(&rfds);
    FD_SET(fileno(stdin),&rfds);

    int r = select(1,&rfds,0,0,&timeout);
    if(r==1) return 1;
    return 0;
}
#endif


class gui {
public:;
    class ffs *ffs;
    void repaint();
    void run();
    void help();
    void search();
    int block;
    char argbuf[512];				// argument we are entering
    char last_search[512];
    int blocks() { return ffs->img->image_blocks;}
    int arg() { if(argbuf[0]) return atoi(argbuf);return 1;}
    gui(class ffs *ffs_) {
	ffs = ffs_;
	block = 0;
	last_search[0] = 0;
    };
};

/* Paint the commands */

void gui::repaint()
{
    unsigned char blockbuf[512];
    char buf[600];

    clear();
    mvprintw(0,0,"Block %6d/%6d",block,blocks());
    mvprintw(LINES-2,0,"sp forward; del back; g goto; / search");
    mvprintw(LINES-1,0,"[0-9] arg");
    
    int y = 2;				// where we are going to put the block number
    if(LINES>20){			// big enough to put in the preceeding block
	if(block>0){
	    memset(blockbuf,0,sizeof(blockbuf));
	    ffs->img->readblock(block-1,blockbuf);
	    img::block2pp(blockbuf,buf);
	    mvaddstr(y,0,buf);
	    if(COLS>80){
		mvprintw(y,66,"[%6d]",block-1);
	    }
	}
	y+=9;

	memset(blockbuf,0,sizeof(blockbuf));
	ffs->img->readblock(block,blockbuf);
	img::block2pp(blockbuf,buf);
	mvaddstr(y,0,buf);
	if(COLS>80){
	    mvprintw(y,66,"[%6d]",block);
	}
	y+=9;

	if(y<LINES+12){
	    memset(blockbuf,0,sizeof(blockbuf));
	    ffs->img->readblock(block+1,blockbuf);
	    img::block2pp(blockbuf,buf);
	    mvaddstr(y,0,buf);
	    if(COLS>80){
		mvprintw(y,66,"[%6d]",block+1);
	    }
	    y+=9;
	}
    }
    refresh();
}

void gui::help()
{
    clear();
    mvprintw(0,0,"Help is not currently available");
    refresh();
    keychar();
}

void gui::search()
{
    char searchbuf[512];

    memset(searchbuf,0,sizeof(searchbuf));
    move(LINES-2,0);			// go to our command line
    clrtoeol();				// clear to end of line
    mvprintw(LINES-2,0,"Search: ");	// ask for a search string
    refresh();
    echo();
    getnstr(searchbuf,sizeof(searchbuf)); // get input
    noecho();

    if(searchbuf[0]==0) strcpy(searchbuf,last_search);
    else strcpy(last_search,searchbuf);

    if(searchbuf[0]){
	for(unsigned int i=block+1;i<blocks();i++){
	    unsigned char blockbuf[513];
	    
	    if(i%10000==0){
		mvprintw(LINES-2,0,"Searching for '%s'  %d/%d...",searchbuf,i,blocks());
		clrtoeol();
		refresh();
	    }
	    
	    ffs->img->readblock(i,blockbuf);
	    blockbuf[512] = 0;
	    if(strstr((char *)blockbuf,searchbuf)){
		block = i;
		return;			// found!
	    }
	}
	mvprintw(LINES-2,0,"NOT FOUND.  Press any key to continue...");
	refresh();
	getch();
    }

}

void gui::run()
{
    do {
	argbuf[0] = 0;
	repaint();

    again:;
	int ch = getch();

	switch(ch){
	case '0': case '1': case'2': case '3': case '4': case '5':
	case '6': case '7': case '8': case '9':
	    if(strlen(argbuf)<sizeof(argbuf-1)){
		char b2[2] = {ch,0};
		strcat(argbuf,b2);
		mvprintw(LINES-3,0,"%s",argbuf);
		refresh();
	    }
	    goto again;

	case '\177':
	case '\010':
	case '-':
	    block -= arg();
	    if(block < 0) block = 0;
	    break;

	case '+':
	case ' ':
	    block += arg();
	    if(block>blocks()){
		block = blocks() - 1;
	    }
	    break;

	case '/':
	    search();
	    break;

	case 'h':
	case '?':
	    help();
	    break;

	case 'q':
	    return;
	}
    } while(1);
}


void gui_main(class ffs *ffs)
{
    initscr(); cbreak(); noecho();
    nonl();
    intrflush(stdscr,FALSE);
    keypad(stdscr, TRUE);
    
    gui g(ffs);
    g.run();

    endwin();
    puts("We hope you enjoyed your experience.");
}
