/*
 * scandir.cpp
 *
 * Recursively scans a FAT32 directory.
 */

#include "StdAfx.h"			// so we can compile under VC6.0
#include "ffs.h"			// bring in the forensic file system
#include "fatdump.h"
#include "getopt.h"


/****************************************************************
 * Directory scanning logic.
 * Two modes of operation:
 * Mode #1 - building block allocation map of level0&1 data.
 * Mode #2 - displaying the contents of all directories.
 */

class recursive_callback : public ffs::dir_callback {
protected:;
    class ffs  *fs;
    char *basedir;			// base directory name
public:;
    fat_dirent *dirs;
    int ndirs;
    recursive_callback(class ffs *fs_,const char *basedir_) {
	fs = fs_;
	basedir = basedir_ ? strdup(basedir_) : 0;
	dirs=(fat_dirent *)malloc(0);ndirs=0;
    }
    ~recursive_callback() {
	if(basedir) free(basedir);
    }
    void dir(class fat_dirent *);
};

void recursive_callback::dir(class fat_dirent *dir) {
    if(opt_v>0) printf("recursive_callback::dir(%s)\n",dir->name());
    dirs = (fat_dirent *)realloc(dirs,sizeof(*dirs)*(ndirs+1));
    dirs[ndirs++] = *dir;
}



/****************************************************************
 ** class buildmap01 --- build a map of the level0 and level 1 files
 ****************************************************************/

class buildmap01 : public recursive_callback {
public:;
    buildmap01(class ffs *fs_,const char *basedir_)
	: recursive_callback(fs_,basedir_) {
    }
    void file(class fat_dirent *);
    void dir(class fat_dirent *);
};

void buildmap01::file(class fat_dirent *file)
{
    if(file->level()==0 ||
       file->level()==1){
	ffs_chain clusters = fs->get_clusterchain(file->DIR_FstClus); // get a clusterchain for the file
	clustermap.annotate(clusters,A_USED01);	// and note that the whole file is used
    }
    recursive_callback::file(file);
}

void buildmap01::dir(class fat_dirent *file)
{
    if(file->level()==0 ||
       file->level()==1){
	ffs_chain clusters = fs->get_clusterchain(file->DIR_FstClus);
	clustermap.annotate(clusters,A_USED01);
	recursive_callback::dir(file);
    }
}

/* map_level01:
 * Build the map of all level01 files.
 * similar to scan_dir, but special purpose.
 */
 
void map_level01(class ffs &fs,const char *basedir,pos loc,int blocks)
{
    if(opt_v>0) printf("map_level01(%s)\n",basedir);

    struct fat_dirent dir(&fs.bpb);	// spare directory entry
    class buildmap01 map01(&fs,basedir);
    
    /* If we are given a cluster to investigate,
     * allow ffs routines to go through it.
     */
    if(loc.is_cluster()){
	sector_t sector = loc.sector(&fs.bpb);
	if(sectormap.test(sector,A_USED01)==0){
	    fs.fwrite_dir(0,loc,0,&dir,&map01,0);
	    clustermap.set(loc.cluster(),A_USED01);
	    sectormap.set(sector,A_USED01);
	}
    }

    /* This code is for bootstrapping from a FAT16 root directory */
    if(loc.is_sector()){
	for(int i=0;i<blocks;i++){
	    sector_t sector = loc.sector(&fs.bpb) + i;
	    if(sectormap.test(i,A_USED01)==0){ // if we haven't used this one yet
		fs.fwrite_dir(0,pos(sector),0,&dir,&map01,0); 
		sectormap.set(i,A_USED01);
	    }
	}
    }
    
    /* In either case, call the sub directories */
    for(int i=0;i<map01.ndirs;i++){
	char *newdir = 0;
	if(basedir){
	    newdir = (char *)malloc(16+strlen(basedir)
				    +strlen((char *)map01.dirs[i].name()));
	    strcpy(newdir,basedir);
	    if(newdir[1]!='\000') strcat(newdir,"/");
	    strcat(newdir,(char *)map01.dirs[i].name());
	}
	
	sector_t cluster = map01.dirs[i].DIR_FstClus;
	map_level01(fs,newdir,pos(cluster,0),-1);	// take length from clusterchain
	if(newdir) free(newdir);
    }
}

/****************************************************************
 ** class scan_dir_callback --- display the files
 ****************************************************************/

class scan_dir_callback : public recursive_callback  {
    bool display_fn;			// display file names
public:;
    scan_dir_callback(class ffs *fs_,const char *basedir_,bool display_fn_)
	: recursive_callback(fs_,basedir_) {
	display_fn = display_fn_;
	nfiles=0;
    }
    ~scan_dir_callback() {
	free(dirs);
    }
    void file(class fat_dirent *);
    void dir(class fat_dirent *);
    int nfiles;
};

void scan_dir_callback::file(class fat_dirent *file)
{
    if(opt_v>0) printf("scan_dir_callback::file(%s)\n",file->name());

    // get the sectorchain for the file
    ffs_chain clusters = fs->get_clusterchain(file->DIR_FstClus); 
    int unique_level2=0;
    if(file->level()==2){
	// note that used for a level2 file
	unique_level2 = clustermap.annotate(clusters,A_USED2); 
    }

    if(opt_no0 && file->level()==0) return; // no level0 files
    if(opt_no1 && file->level()==1) return; // no level1 files
    if(opt_no2 && file->deleted) return;    // no level2 files
    nfiles++;

    /* Determine how much of the file is usable */
    int status=0;
    if(file->level()==0 || file->level()==1){
	status = F_COMPLETE;		// by definition;
    }
    if(file->level()==2){
	int level01_clusters = clustermap.union_count(clusters,A_USED01);
	
	if(clusters.count()==level01_clusters){
	    status = F_NONE;		// all clusters are mapped to a level01 file
	}
	else {				// we have some of the clusters
	    if(unique_level2==0){	// none unique
		status = F_NONE;
	    }
	    status = F_PARTIAL;		// assume that it is partial
	    if(unique_level2==clusters.count()){
		/* All of our clusters are unique. See if we have enough
		 * to make a whole file...
		 */
		if(unique_level2 * fs->bpb.BPB_BytsPerSec * fs->bpb.BPB_SecPerClus > file->DIR_FileSize){
		    status = F_COMPLETE; // yes! got enough bytes!
		} else {
		    /* Incomplete */
		}
	    }
	}
    }
    
    times.add(mktime(&file->mtime));

    if(opt_extensions){
	struct extension_db::ent e;
	memset((void *)&e,0,sizeof(e));
	if(status==F_COMPLETE){
	    e.partial[file->level()]++;
	}
	if(status==F_PARTIAL){
	    e.complete[file->level()]++;
	}
	extensions.add(file->name(),e);
    }
    
    if(display_fn){				// if display file names...
	if(opt_lsl){
	    char buf[64];

	    fputs(file->attr2ascii(),stdout);
	    putchar(' ');
	    fprintf(stdout,"%10s {%8d} ",
		    ffs::itocomma(buf,file->DIR_FileSize),
		    file->DIR_FstClus);

	    /**************** report any annotated blocks ****************/
	    switch(status){
	    case F_NONE:    putchar('!');break;
	    case F_COMPLETE:putchar(' ');break;
	    case F_PARTIAL: putchar('?');break;
	    }
	    
	    putchar(' ');
	}

	putchar(' ');
	fputs(file->mtime2ascii(),stdout);
	putchar(' ');

	if(basedir && basedir[1]) fputs(basedir,stdout);
	if(basedir) putchar('/');
	fputs(file->name(),stdout);
	if(opt_showLevel) fprintf(stdout," <%d>",file->level());
	if(opt_showClusterChain){
	    putchar(' ');
	    char buf[32];
	    fs->fat0.get_clusterchain(file->DIR_FstClus,buf,sizeof(buf));
	    fputs(buf,stdout);
	}
	putchar('\n');
    }
}

void scan_dir_callback::dir(class fat_dirent *dir) {
    /* Note: don't filter out level0 dirs, as we would miss level2 files in level0 dirs! */
    if(opt_no1 && dir->level()==1) return; // no level1 dirs
    if(opt_no2 && dir->deleted) return; // no level2 files
    recursive_callback::dir(dir);	// call to super
}

/* Scan a directory on the disk.
 * Returns the number of entries that were found.
 * Designed to be called recursively.
 */
int scan_dir(class ffs &fs,const char *basedir,pos loc,int blocks)
{
    if(opt_v>0) printf("scan_dir(,%s,,%d)\n",basedir,blocks);


    int flags = 0;
    struct fat_dirent dir(&fs.bpb);	// spare directory entry
    int i;
    int entries=0;
    bool display_filenames = opt_dir==0 && opt_q==0;
    FILE *outfile = opt_dir ? stdout : 0;	

    class scan_dir_callback scb(&fs,basedir,display_filenames);

    if(opt_v) flags |= ffs::DIR_SHOWLDIR;

    /* If we are given a cluster to investigate, allow ffs routines to go through it.
     */
    if(loc.is_cluster()){
	sector_t sector = loc.sector(&fs.bpb);
	if(sectormap.test(sector,A_USED_DIR)==0){
	    if(opt_v) printf("== cluster %d (starts at sector %d) ==\n",
			     loc.cluster(), sector);
	    fs.fwrite_dir(outfile,loc,0,&dir,&scb,flags);
	    sectormap.set(sector,A_USED_DIR);
	}
    }
    /*
     * If we are given a specific sector to investigate, just check it
     * This is protected against loops.
     */
    /* This code is for bootstrapping from a FAT16 root directory */
    if(loc.is_sector()){
	for(i=0;i<blocks;i++){
	    sector_t sector = loc.sector(&fs.bpb) + i;
	    if(sectormap.test(i,A_USED_DIR)==0){ // if we haven't used this one yet
		if(opt_v) printf("== sector %d ==\n",sector);
		fs.fwrite_dir(outfile,pos(sector),0,&dir,&scb,flags); 
		sectormap.set(i,A_USED_DIR);
	    }
	}
    }

    /* In either case, if we do not not want level 3 data, get it */

    if(!opt_no3){
	for(i=0;i<scb.ndirs;i++){
	    char *newdir = 0;
	    if(basedir){
		newdir = (char *)malloc(16+strlen(basedir)+strlen((char *)scb.dirs[i].name()));
		strcpy(newdir,basedir);
		if(newdir[1]!='\000') strcat(newdir,"/");
		strcat(newdir,(char *)scb.dirs[i].name());
	    }

	    sector_t cluster = scb.dirs[i].DIR_FstClus;
	    entries += scan_dir(fs,newdir,pos(cluster,0),-1);	// take length from clusterchain
	    if(newdir) free(newdir);
	}
    }
    entries += scb.nfiles + scb.ndirs;
    return entries;
}

